@@ -42,8 +42,12 @@ typedef struct
 {
         GdmSettings  *settings;
         GCancellable *cancellable;
-        GSubprocess  *bus_subprocess;
-        char         *bus_address;
+
+        GSubprocess     *bus_subprocess;
+        GDBusConnection *bus_connection;
+        char            *bus_address;
+
+        char           **environment;
 
         GSubprocess  *session_subprocess;
         char         *session_command;
@@ -89,13 +93,13 @@ static gboolean
 spawn_bus (State        *state,
            GCancellable *cancellable)
 {
+        GDBusConnection     *bus_connection = NULL;
         GPtrArray           *arguments = NULL;
         GSubprocessLauncher *launcher = NULL;
         GSubprocess         *subprocess = NULL;
         GInputStream        *input_stream = NULL;
         GDataInputStream    *data_stream = NULL;
         GError              *error = NULL;
-        const char          *bus_env = NULL;
         char                *bus_address_fd_string = NULL;
         char                *bus_address = NULL;
         gsize                bus_address_size;
@@ -106,10 +110,13 @@ spawn_bus (State        *state,
 
         g_debug ("Running session message bus");
 
-        bus_env = g_getenv ("DBUS_SESSION_BUS_ADDRESS");
-        if (bus_env != NULL) {
+        bus_connection = g_bus_get_sync (G_BUS_TYPE_SESSION,
+                                         cancellable,
+                                         NULL);
+
+        if (bus_connection != NULL) {
                 g_debug ("session message bus already running, not starting another one");
-                state->bus_address = g_strdup (bus_env);
+                state->bus_connection = bus_connection;
                 return TRUE;
         }
 
@@ -174,6 +181,20 @@ spawn_bus (State        *state,
                                  on_bus_finished,
                                  state);
 
+        bus_connection = g_dbus_connection_new_for_address_sync (state->bus_address,
+                                                                 G_DBUS_CONNECTION_FLAGS_AUTHENTICATION_CLIENT |
+                                                                 G_DBUS_CONNECTION_FLAGS_MESSAGE_BUS_CONNECTION,
+                                                                 NULL,
+                                                                 cancellable,
+                                                                 &error);
+
+        if (bus_connection == NULL) {
+                g_debug ("could not open connection to session bus: %s",
+                         error->message);
+                goto out;
+        }
+
+        state->bus_connection = bus_connection;
         is_running = TRUE;
 out:
         g_clear_object (&data_stream);
@@ -184,6 +205,38 @@ out:
         return is_running;
 }
 
+static gboolean
+import_environment (State        *state,
+                    GCancellable *cancellable)
+{
+        g_autoptr(GVariant) reply = NULL;
+        g_autoptr(GVariant) environment_variant = NULL;
+        g_autoptr(GError)   error = NULL;
+
+        reply = g_dbus_connection_call_sync (state->bus_connection,
+                                             "org.freedesktop.systemd1",
+                                             "/org/freedesktop/systemd1",
+                                             "org.freedesktop.DBus.Properties",
+                                             "Get",
+                                             g_variant_new ("(ss)",
+                                                            "org.freedesktop.systemd1.Manager",
+                                                            "Environment"),
+                                             NULL,
+                                             G_DBUS_CALL_FLAGS_NONE,
+                                             -1, cancellable, &error);
+
+        if (reply == NULL) {
+                g_debug ("could not fetch environment: %s", error->message);
+                return FALSE;
+        }
+
+        g_variant_get (reply, "(v)", &environment_variant);
+
+        state->environment = g_variant_dup_strv (environment_variant, NULL);
+
+        return TRUE;
+}
+
 static void
 on_session_finished (GSubprocess  *subprocess,
                      GAsyncResult *result,
@@ -241,7 +294,30 @@ spawn_session (State        *state,
         }
 
         launcher = g_subprocess_launcher_new (G_SUBPROCESS_FLAGS_NONE);
-        g_subprocess_launcher_setenv (launcher, "DBUS_SESSION_BUS_ADDRESS", state->bus_address, TRUE);
+
+        if (state->environment != NULL) {
+                size_t i;
+
+                for (i = 0; state->environment[i] != NULL; i++) {
+                        g_auto(GStrv) environment_entry = NULL;
+
+                        if (state->environment[i] == '\0') {
+                                continue;
+                        }
+
+                        environment_entry = g_strsplit (state->environment[i], "=", 2);
+
+                        if (environment_entry[0] == NULL || environment_entry[1] == NULL) {
+                                continue;
+                        }
+
+                        g_subprocess_launcher_setenv (launcher, environment_entry[0], environment_entry[1], FALSE);
+                }
+        }
+
+        if (state->bus_address != NULL) {
+                g_subprocess_launcher_setenv (launcher, "DBUS_SESSION_BUS_ADDRESS", state->bus_address, TRUE);
+        }
         subprocess = g_subprocess_launcher_spawnv (launcher,
                                                    (const char * const *) argv,
                                                    &error);
@@ -344,7 +420,9 @@ clear_state (State **out_state)
         State *state = *out_state;
 
         g_clear_object (&state->cancellable);
+        g_clear_object (&state->bus_connection);
         g_clear_object (&state->session_subprocess);
+        g_clear_pointer (&state->environment, g_strfreev);
         g_clear_pointer (&state->main_loop, g_main_loop_unref);
         *out_state = NULL;
 }
@@ -402,7 +480,7 @@ main (int    argc,
         ret = gdm_settings_direct_init (state->settings, DATADIR "/gdm/gdm.schemas", "/");
 
         if (!ret) {
-                g_printerr ("Unable to initialize settings");
+                g_printerr ("Unable to initialize settings\n");
                 exit_status = EX_DATAERR;
                 goto out;
         }
@@ -420,15 +498,17 @@ main (int    argc,
         ret = spawn_bus (state, state->cancellable);
 
         if (!ret) {
-                g_printerr ("Unable to run session message bus");
+                g_printerr ("Unable to run session message bus\n");
                 exit_status = EX_SOFTWARE;
                 goto out;
         }
 
+        import_environment (state, state->cancellable);
+
         ret = spawn_session (state, state->cancellable);
 
         if (!ret) {
-                g_printerr ("Unable to run session");
+                g_printerr ("Unable to run session\n");
                 exit_status = EX_SOFTWARE;
                 goto out;
         }
@@ -436,7 +516,7 @@ main (int    argc,
         ret = register_display (state, state->cancellable);
 
         if (!ret) {
-                g_printerr ("Unable to register display with display manager");
+                g_printerr ("Unable to register display with display manager\n");
                 exit_status = EX_SOFTWARE;
                 goto out;
         }
