@@ -49,8 +49,11 @@ typedef struct
         char         *auth_file;
         char         *display_name;
 
-        GSubprocess  *bus_subprocess;
-        char         *bus_address;
+        GSubprocess     *bus_subprocess;
+        GDBusConnection *bus_connection;
+        char            *bus_address;
+
+        char           **environment;
 
         GSubprocess  *session_subprocess;
         char         *session_command;
@@ -373,30 +376,16 @@ static gboolean
 update_bus_environment (State        *state,
                         GCancellable *cancellable)
 {
-        GDBusConnection     *connection = NULL;
         GVariantBuilder      builder;
         GVariant            *reply = NULL;
         GError              *error = NULL;
         gboolean             environment_updated = FALSE;
 
-        connection = g_dbus_connection_new_for_address_sync (state->bus_address,
-                                                             G_DBUS_CONNECTION_FLAGS_AUTHENTICATION_CLIENT |
-                                                             G_DBUS_CONNECTION_FLAGS_MESSAGE_BUS_CONNECTION,
-                                                             NULL,
-                                                             cancellable,
-                                                             &error);
-
-        if (connection == NULL) {
-                g_debug ("could not open connection to session bus: %s",
-                         error->message);
-                goto out;
-        }
-
         g_variant_builder_init (&builder, G_VARIANT_TYPE ("a{ss}"));
         g_variant_builder_add (&builder, "{ss}", "DISPLAY", state->display_name);
         g_variant_builder_add (&builder, "{ss}", "XAUTHORITY", state->auth_file);
 
-        reply = g_dbus_connection_call_sync (connection,
+        reply = g_dbus_connection_call_sync (state->bus_connection,
                                              "org.freedesktop.DBus",
                                              "/org/freedesktop/DBus",
                                              "org.freedesktop.DBus",
@@ -405,7 +394,7 @@ update_bus_environment (State        *state,
                                                             g_variant_builder_end (&builder)),
                                              NULL,
                                              G_DBUS_CALL_FLAGS_NONE,
-                                             -1, NULL, &error);
+                                             -1, cancellable, &error);
 
         if (reply == NULL) {
                 g_debug ("could not update activation environment: %s", error->message);
@@ -417,7 +406,6 @@ update_bus_environment (State        *state,
         environment_updated = TRUE;
 
 out:
-        g_clear_object (&connection);
         g_clear_error (&error);
 
         return environment_updated;
@@ -427,13 +415,13 @@ static gboolean
 spawn_bus (State        *state,
            GCancellable *cancellable)
 {
+        GDBusConnection     *bus_connection = NULL;
         GPtrArray           *arguments = NULL;
         GSubprocessLauncher *launcher = NULL;
         GSubprocess         *subprocess = NULL;
         GInputStream        *input_stream = NULL;
         GDataInputStream    *data_stream = NULL;
         GError              *error = NULL;
-        const char          *bus_env = NULL;
         char                *bus_address_fd_string;
         char                *bus_address = NULL;
         gsize                bus_address_size;
@@ -444,10 +432,13 @@ spawn_bus (State        *state,
 
         g_debug ("Running session message bus");
 
-        bus_env = g_getenv ("DBUS_SESSION_BUS_ADDRESS");
-        if (bus_env != NULL) {
+        bus_connection = g_bus_get_sync (G_BUS_TYPE_SESSION,
+                                         cancellable,
+                                         NULL);
+
+        if (bus_connection != NULL) {
                 g_debug ("session message bus already running, not starting another one");
-                state->bus_address = g_strdup (bus_env);
+                state->bus_connection = bus_connection;
                 return TRUE;
         }
 
@@ -513,6 +504,22 @@ spawn_bus (State        *state,
                                  on_bus_finished,
                                  state);
 
+
+        bus_connection = g_dbus_connection_new_for_address_sync (state->bus_address,
+                                                                 G_DBUS_CONNECTION_FLAGS_AUTHENTICATION_CLIENT |
+                                                                 G_DBUS_CONNECTION_FLAGS_MESSAGE_BUS_CONNECTION,
+                                                                 NULL,
+                                                                 cancellable,
+                                                                 &error);
+
+        if (bus_connection == NULL) {
+                g_debug ("could not open connection to session bus: %s",
+                         error->message);
+                goto out;
+        }
+
+        state->bus_connection = bus_connection;
+
         is_running = TRUE;
 out:
         g_clear_object (&data_stream);
@@ -523,6 +530,38 @@ out:
         return is_running;
 }
 
+static gboolean
+import_environment (State        *state,
+                    GCancellable *cancellable)
+{
+        g_autoptr(GVariant) reply = NULL;
+        g_autoptr(GVariant) environment_variant = NULL;
+        g_autoptr(GError)   error = NULL;
+
+        reply = g_dbus_connection_call_sync (state->bus_connection,
+                                             "org.freedesktop.systemd1",
+                                             "/org/freedesktop/systemd1",
+                                             "org.freedesktop.DBus.Properties",
+                                             "Get",
+                                             g_variant_new ("(ss)",
+                                                            "org.freedesktop.systemd1.Manager",
+                                                            "Environment"),
+                                             NULL,
+                                             G_DBUS_CALL_FLAGS_NONE,
+                                             -1, cancellable, &error);
+
+        if (reply == NULL) {
+                g_debug ("could not fetch environment: %s", error->message);
+                return FALSE;
+        }
+
+        g_variant_get (reply, "(v)", &environment_variant);
+
+        state->environment = g_variant_dup_strv (environment_variant, NULL);
+
+        return TRUE;
+}
+
 static void
 on_session_finished (GSubprocess  *subprocess,
                      GAsyncResult *result,
@@ -573,7 +612,30 @@ spawn_session (State        *state,
 
         g_subprocess_launcher_setenv (launcher, "DISPLAY", state->display_name, TRUE);
         g_subprocess_launcher_setenv (launcher, "XAUTHORITY", state->auth_file, TRUE);
-        g_subprocess_launcher_setenv (launcher, "DBUS_SESSION_BUS_ADDRESS", state->bus_address, TRUE);
+
+        if (state->environment != NULL) {
+                size_t i;
+
+                for (i = 0; state->environment[i] != NULL; i++) {
+                        g_auto(GStrv) environment_entry = NULL;
+
+                        if (state->environment[i] == '\0') {
+                                continue;
+                        }
+
+                        environment_entry = g_strsplit (state->environment[i], "=", 2);
+
+                        if (environment_entry[0] == NULL || environment_entry[1] == NULL) {
+                                continue;
+                        }
+
+                        g_subprocess_launcher_setenv (launcher, environment_entry[0], environment_entry[1], FALSE);
+                }
+        }
+
+        if (state->bus_address != NULL) {
+                g_subprocess_launcher_setenv (launcher, "DBUS_SESSION_BUS_ADDRESS", state->bus_address, TRUE);
+        }
 
         vt = g_getenv ("XDG_VTNR");
 
@@ -712,8 +774,10 @@ clear_state (State **out_state)
         State *state = *out_state;
 
         g_clear_object (&state->cancellable);
+        g_clear_object (&state->bus_connection);
         g_clear_object (&state->session_subprocess);
         g_clear_object (&state->x_subprocess);
+        g_clear_pointer (&state->environment, g_strfreev);
         g_clear_pointer (&state->auth_file, g_free);
         g_clear_pointer (&state->display_name, g_free);
         g_clear_pointer (&state->main_loop, g_main_loop_unref);
@@ -808,6 +872,8 @@ main (int    argc,
                 goto out;
         }
 
+        import_environment (state, state->cancellable);
+
         ret = update_bus_environment (state, state->cancellable);
 
         if (!ret) {
